/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include "is_types.h"
#include "isam.h"

#ifdef	BSD
#define	memcmp(a,b,n)	bcmp(a,b,n)
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)db_update.c	1.8	11:48:36	2/7/92";
#endif

long	isam_set_ticket();
long	__isam_lookup();

extern	int	db_errno;

/*
 * NAME:
 *	__db_update_record
 *
 * NOTES:
 *	This routine relies on index 1 being the primary record number
 *	index.  A lookup will be made against that index to determine
 *	if the current record is still valid.
 */

int
__db_update_record (file, data, validate)
file_t	*file;
char	*data;
int	validate;
{
	int	i;
	off_t	record = -1;
	sort_key_t *key,
		*keys = file->f_index->i_keys;
	field_t *fields = file->f_schema->s_fields;
	int	nfields = file->f_schema->s_field_count;
	int	was_locked = file->f_locked;
	long	ticket;
	char	old_data[MAXRECSIZE];
	char	old_key[MAXBLOCKSIZE];
	char	new_key[MAXBLOCKSIZE];

	/*
	 * Start off with a write-lock.  The file will be modified quite
	 * extensively, so it must be locked before anything significant
	 * is done.
	 */

	if (was_locked != F_WRLCK)
		if ((ticket = isam_set_ticket (file)) == -1)
			return -1;

	/*
	 * Locate the current record in the primary index.  This acts
	 * as a sanity check to insure the record really exists.  This
	 * may be suppressed incase there is no primary index (which
	 * the data dictionary doesn't have, for example).
	 */

	if ((record = file->f_current_record) <= 0) {
		db_errno = DB_NO_RECORD;
		goto fail;
	}
	if (validate) {
		if (__isam_lookup (file, &keys[1], 1, (void *) &record,
				IS_EQUAL, IS_UNIQUE) != record) {
			db_errno = DB_NO_RECORD;
			file->f_current_record = -1;
			goto fail;
		}
	}
	if (isam_read_data (file, record, old_data))
		goto fail;

	/*
	 * See if there is a serial number field in this file.  If there
	 * is, we will have to preserve the serial number for this record.
	 * We do this by getting the old serial number out of the old_data
	 * and copying it into the user's buffer.
	 */

	for (i = 0;i < nfields;i++) {
		if (fields[i].f_type == T_SERIAL) {
			memcpy (data + fields[i].f_offset,
				old_data + fields[i].f_offset,
				sizeof file->f_serial);

			break;
		}
	}

	/*
	 * Check all of the keys for the record.  If the record has
	 * changed a particular key, add the new key and delete the
	 * old.  Don't delete any of the old ones until all of the
	 * new ones have been added ...
	 */

	for (i = 1;i <= file->f_index->i_key_count;i++) {
		if (i == file->f_current_index)
			continue;

		key = &keys[i];
		__isam_buf2key (file, i, old_data, record, old_key);
		__isam_buf2key (file, i, data, record, new_key);
		if (memcmp (old_key, new_key, key->k_key_size) == 0)
			continue;

		if (isam_add (file, i, new_key, record))
			goto undo_new;
	}
	if (i = file->f_current_index) {
		key = &keys[i];
		__isam_buf2key (file, i, old_data, record, old_key);
		__isam_buf2key (file, i, data, record, new_key);
		if (memcmp (old_key, new_key, key->k_key_size) != 0)
			if (isam_add (file, i, new_key, record))
				goto undo_new;
	}

	/*
	 * I now have added all of the keys needed to find this thing,
	 * so I might as well write it out.
	 */

	if (isam_write_data (file, record, data))
		goto undo_new;

	/*
	 * Now I have to go back and delete all of the keys for
	 * the old record.
	 */

	for (i = 2;i <= file->f_index->i_key_count;i++) {
		if (i == file->f_current_index)
			continue;

		key = &keys[i];
		__isam_buf2key (file, i, old_data, record, old_key);
		__isam_buf2key (file, i, data, record, new_key);
		if (memcmp (old_key, new_key, key->k_key_size) == 0)
			continue;

		if (isam_delete (file, i, old_key, record))
			goto undo_old;
	}
	if (i = file->f_current_index) {
		key = &keys[i];
		__isam_buf2key (file, i, old_data, record, old_key);
		__isam_buf2key (file, i, data, record, new_key);
		if (memcmp (old_key, new_key, key->k_key_size) != 0) {
			if (isam_delete (file, i, old_key, record))
				goto undo_old;
		}
		if (__isam_lookup (file, &keys[i], i, (void *) new_key,
				IS_EQUAL, IS_UNIQUE) == (off_t) -1)
			goto fail;
	}

	/*
	 * Finally I am done screwing with the index so it is safe
	 * to unlock the file.
	 */

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return 0;

fail:
	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return -1;

undo_new:
	for (i = 2;i <= file->f_index->i_key_count;i++)
		(void) __db_delete_key (file, i, data, record);

	goto fail;

undo_old:
	for (i = 2;i <= file->f_index->i_key_count;i++)
		(void) __db_delete_key (file, i, data, record);

	goto fail;
}

int
db_update_record (dbm, filenum, buf)
dbm_t	*dbm;
int	filenum;
VOIDP	*buf;
{
	file_t	*file;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	file = &dbm->d_files[filenum];
	if (file->f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	if (file->f_index == 0)
		return (db_errno = DB_BAD_DB), -1;

	return __db_update_record (file, (char *) buf, 1);
}
